#include "ofTrueTypeFont.h"
//--------------------------


#ifdef TARGET_WIN32
	#include <ft2build.h>
	#include <freetype/freetype.h>
	#include <freetype/ftglyph.h>
	#include <freetype/ftoutln.h>
	#include <freetype/fttrigon.h>
#endif
#ifdef TARGET_OSX
	#include "ft2build.h"
	#include "freetype.h"
	#include "ftglyph.h"
	#include "ftoutln.h"
	#include "fttrigon.h"
#endif



//------------------------------------------------------------------
ofTrueTypeFont::ofTrueTypeFont(){
	bLoadedOk = false;
}

//------------------------------------------------------------------
void ofTrueTypeFont::loadFont(const char *filename, int fontsize){
	// load anti-aliased, non-full character set:
	loadFont(filename, fontsize, true, false);
}

//------------------------------------------------------------------
void ofTrueTypeFont::loadFont(const char *filename, int fontsize, bool _bAntiAliased, bool _bFullCharacterSet){
	
	bLoadedOk 			= false;
	bAntiAlised 		= _bAntiAliased;
	bFullCharacterSet 	= _bFullCharacterSet;
	
	//--------------- load the library and typeface
	FT_Library library;
	if (FT_Init_FreeType( &library )){
		printf("PROBLEM WITH FT lib \n");
		return;
	}

	FT_Face face;
	if (FT_New_Face( library, filename, 0, &face )) {
		return;
	}
	
	FT_Set_Char_Size( face,fontsize << 6,fontsize << 6,96,96);
	lineHeight = fontsize * 1.43f;
	
	//------------------------------------------------------
	//kerning would be great to support:
	//printf("FT_HAS_KERNING ? %i \n", FT_HAS_KERNING(face));
	//------------------------------------------------------
	
	nCharacters = bFullCharacterSet ? 256 : 128 - NUM_CHARACTER_TO_START;
	
	
	//--------------- initialize character info and textures
	cps       = new charProps[nCharacters];
	texNames  = new GLuint[nCharacters];
	glGenTextures(nCharacters, texNames);
	
	//--------------------- load each char -----------------------
	for (int i = 0 ; i < nCharacters; i++){
	
		//------------------------------------------ anti aliased or not:
		if (bAntiAlised == true){
			if(FT_Load_Glyph( face, FT_Get_Char_Index( face, (unsigned char)(i+NUM_CHARACTER_TO_START) ), FT_LOAD_DEFAULT )){
				printf("error with FT_Load_Glyph %i \n", i);
			}
			FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);
		} else {
			if(FT_Load_Glyph( face, FT_Get_Char_Index( face, (unsigned char)(i+NUM_CHARACTER_TO_START) ), FT_LOAD_DEFAULT )){
				printf("error with FT_Load_Glyph%i \n", i);
			}
			FT_Render_Glyph(face->glyph, FT_RENDER_MODE_MONO); 
		}
		//------------------------------------------

		FT_Bitmap& bitmap= face->glyph->bitmap;


		// 1 pixel border around the glyph
		int border = 2;
		
		// prepare the texture:
		
		int width  = ofNextPow2( bitmap.width + border);
		int height = ofNextPow2( bitmap.rows  + border);
		
		// ------------------------- this is fixing a bug with small type
		// ------------------------- appearantly, opengl has trouble with 
		// ------------------------- width or height textures of 1, so we
		// ------------------------- we just set it to 2...
		if (width == 1) width = 2;
		if (height == 1) height = 2;
		// -------------------------
		
		// info about the character:
		cps[i].value 			= i;
		cps[i].height 			= face->glyph->bitmap_top;
		cps[i].width 			= face->glyph->bitmap.width;
		cps[i].setWidth 		= face->glyph->advance.x >> 6;
		cps[i].topExtent 		= face->glyph->bitmap.rows;
		cps[i].leftExtent		= face->glyph->bitmap_left;
		// texture internals
		cps[i].tTex				= (float)bitmap.width / (float)width;
		cps[i].vTex				= (float)bitmap.rows /  (float)height;
		
		cps[i].xOff				= ((1.0+(float)bitmap.width) / (float)width)  - cps[i].tTex;
		cps[i].yOff				= ((1.0+(float)bitmap.rows)  / (float)height) - cps[i].vTex;
		
		/* sanity check:
		printf("%i %i %i %i %i %i \n", 
		cps[i].value ,
		cps[i].height ,
		cps[i].width 	,
		cps[i].setWidth 	,
		cps[i].topExtent ,
		cps[i].leftExtent	);
		*/
		
		// Allocate Memory For The Texture Data.
		unsigned char* expanded_data = new unsigned char[ 2 * width * height];

		//-------------------------------- clear data:
		for(int j=0; j <height;j++) {
			for(int k=0; k < width; k++){
				expanded_data[2*(k+j*width)  ] = 0; 
				expanded_data[2*(k+j*width)+1] = 0;
			}
		}


		if (bAntiAlised == true){
			//-----------------------------------	
			for(int j=1; j <height-1;j++) {
				for(int k=1; k < width-1; k++){
					expanded_data[2*(k+j*width)] = 
						(k-1>=bitmap.width || j-1>=bitmap.rows) ?
						0 : 255;
					expanded_data[2*(k+j*width)+1] = 
						(k-1>=bitmap.width || j-1>=bitmap.rows) ?
						0 : bitmap.buffer[k-1 + bitmap.width*(j-1)];
				}
			}
			//-----------------------------------
		} else {
			//-----------------------------------
			// true type packs monochrome info in a 
			// 1-bit format, hella funky
			// here we unpack it:
			unsigned char *src =  bitmap.buffer;
			for(int j=0; j <bitmap.rows;j++) {
				unsigned char b;
				unsigned char *bptr =  src;
				for(int k=0; k < bitmap.width ; k++){
					expanded_data[2*((k+1)+(j+1)*width)] = 255;
					if (k%8==0){ b = (*bptr++);}
					expanded_data[2*((k+1)+(j+1)*width) + 1] =  
						b&0x80 ? 255 : 0;
				    b <<= 1;
				}
				src += bitmap.pitch;
			}
			//-----------------------------------
		}
		

		//Now we just setup some texture paramaters.
    	
    	
    	glBindTexture( GL_TEXTURE_2D, texNames[i]);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   		if (bAntiAlised == true){   			
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		} else {
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		}
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
		//Here we actually create the texture itself, notice
		//that we are using GL_LUMINANCE_ALPHA to indicate that
		//we are using 2 channel data.
		bool b_use_mipmaps = false;  // FOR now this is fixed to false, could be an option, left in for legacy...
		if (b_use_mipmaps){
   			gluBuild2DMipmaps(
   				GL_TEXTURE_2D, GL_LUMINANCE_ALPHA, width, height,
            	GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expanded_data);
		} else {
	    	glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA, width, height,
			   0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expanded_data );
		}


		//With the texture created, we don't need to expanded data anymore
    	delete [] expanded_data;
   }
   
	// ------------- close the library and typeface
	FT_Done_Face(face);
	FT_Done_FreeType(library);
  	bLoadedOk = true;
}

//-----------------------------------------------------------
int ofTrueTypeFont::ofNextPow2 ( int a )
{
	int rval=1;
	while(rval<a) rval<<=1;
	return rval;
}


//-----------------------------------------------------------
void ofTrueTypeFont::setLineHeight(float _newLineHeight) {
	lineHeight = _newLineHeight;
}

//-----------------------------------------------------------
float ofTrueTypeFont::getLineHeight(){
	return lineHeight;
}




//-----------------------------------------------------------
void ofTrueTypeFont::drawChar(int c, float x, float y) {


	//----------------------- error checking
	if (!bLoadedOk){ 
		printf("Error : font not allocated -- line %d in %s\n", __LINE__,__FILE__);
		return; 
	}
	
	if (c >= nCharacters){
		//printf("Error : char (%i) not allocated -- line %d in %s\n", (c + NUM_CHARACTER_TO_START), __LINE__,__FILE__);
		return; 
	}
	//-----------------------

	int cu = c;

	float height  = cps[cu].height;
	float bwidth  = cps[cu].width;
	float top     = cps[cu].topExtent - cps[cu].height;
	float lextent = cps[cu].leftExtent;

	float t = cps[cu].tTex;
	float v = cps[cu].vTex;

	float xOff	= cps[cu].xOff;
	float yOff  = cps[cu].yOff;

	if (glIsTexture(texNames[cu])) {	
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texNames[cu]);
		glNormal3f(0, 0, 1);
		glBegin(GL_QUADS);
			glTexCoord2d(xOff,yOff);			glVertex2f(x + lextent, y + height);
			glTexCoord2d(xOff,v+yOff);			glVertex2f(x + lextent, y-top);
			glTexCoord2d(t+xOff,v+yOff);		glVertex2f(x + lextent + bwidth,y-top);
			glTexCoord2d(t+xOff,yOff);			glVertex2f(x + + lextent + bwidth,y + height);
		glEnd();
		glDisable(GL_TEXTURE_2D);
	} else {
		//let's add verbosity levels somewhere...
		//this error, for example, is kind of annoying to see
		//all the time:
		//printf("Warning : texture not bound for character -- line %d in %s\n", __LINE__,__FILE__);   
	}

}


//-----------------------------------------------------------
float ofTrueTypeFont::stringWidth(char *c) {
    
    
     float 	width 		= 0;
    int 	index 		= 0;
    float 	widthTemp 	= 0;
    bool 	bMultiLine 	= false;
    
    while (c[index] != 0) { 
		int cy = (unsigned char)c[index] - NUM_CHARACTER_TO_START;
 	    if (cy < nCharacters){ 			// full char set or not? 	    
	       if (c[index] == '\n') {
				width = MAX(width, widthTemp);
				widthTemp = 0;
	      } else if (c[index] == ' ') {
	     		int cy = (int)'p' - NUM_CHARACTER_TO_START;
	      		widthTemp += cps[cy].width; 
	      } else {
				widthTemp += cps[cy].setWidth;
		  }
	  	}
    	index++;
    }
    
    width = MAX(width, widthTemp);
    return width;
    
}


//-----------------------------------------------------------
float ofTrueTypeFont::stringHeight(char *c) {
	
	float 	height 	= 0;
    int 	index 	= 0;
    float 	maxHeight = 0;
    bool 	bFirstLineOfMultiLine = true;
    
    while (c[index] != 0) { 
		int cy = (unsigned char)c[index] - NUM_CHARACTER_TO_START;
 	    if (cy < nCharacters){ 			// full char set or not? 	    
	      if (c[index] == '\n') {
	      		if (bFirstLineOfMultiLine == true){
	      			bFirstLineOfMultiLine = false;
	      			height += maxHeight;
	      		} else {
	      			height += lineHeight; 
	      		}
				maxHeight = 0;
	      } else if (c[index] == ' ') {
	      		 
	      } else {
	      		float ht = cps[cy].height;
	      		if (ht > maxHeight) maxHeight = ht;
		  }
	  	}
    	index++;
    }
    
    height += maxHeight;
    return height;
}



//=====================================================================
void ofTrueTypeFont::drawString(char *c, float x, float y) {
    
    
    
    
    if (!bLoadedOk){ 
    	printf("Error : font not allocated -- line %d in %s\n", __LINE__,__FILE__);
    	return; 
    };
    
    // we need transparency to draw text, but we don't know
    // if that is set up in outside of this function 
    // we "pushAttrib", turn on alpha and "popAttrib"
    // http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/pushattrib.html
    
    // **** note ****
    // I have read that pushAttrib() is slow, if used often, 
    // maybe there is a faster way to do this?
    // ie, check if blending is enabled, etc...
    // glIsEnabled().... glGet()...  
    // http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/get.html
    // **************
    
    // (a) record the current "alpha state, blend func, etc"
    glPushAttrib(GL_COLOR_BUFFER_BIT);
    // (b) enable our regular ALPHA blending!
    glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    
    float startX = x;
    int index = 0;
   
    while (c[index] != 0) { 
		int cy = (unsigned char)c[index] - NUM_CHARACTER_TO_START;
 	    if (cy < nCharacters){ 			// full char set or not? 	    
	      if (c[index] == '\n') {
				x = startX;
				int cy = (int)'M' - NUM_CHARACTER_TO_START;
				y -= lineHeight; 
	      } else if (c[index] == ' ') {
	      		 int cy = (int)'p' - NUM_CHARACTER_TO_START;
	      		 x += cps[cy].width; 
	      } else {
				drawChar(cy, x, y);
				x += cps[cy].setWidth;
		  }
	  	}
    	index++;
    }
    
    // (c) return back to the way things were (with blending, blend func, etc)
    glPopAttrib();
    
}



